@@ -1,41 +1,56 @@
 """DB-SIG compliant module for communicating with MS SQL servers"""
-
-__author__ = 'Mikhail Denisenko <denisenkom@gmail.com>'
-__version__ = '1.8.2'
-
 from collections import deque
-import logging
-import six
+import datetime
+import errno
+import keyword
 import os
 import re
-import keyword
-from six.moves import xrange
-from . import lcid
-from datetime import date, datetime, time
-from . import tz
+import six
 import socket
-import errno
 import uuid
 import warnings
-from .tds import *
-from .tds import (
-    Error, LoginError, DatabaseError,
-    InterfaceError, TimeoutError, OperationalError,
-    TDS_PENDING, TDS74,
-    TDS_ENCRYPTION_OFF, TDS_ODBC_ON, SimpleLoadBalancer,
-    IS_TDS7_PLUS,
-    _TdsSocket, tds7_get_instances, ClosedConnectionError,
-    SP_EXECUTESQL, Column, _create_exception_by_message,
-)
 import weakref
 
+from six.moves import xrange
+
+from pytds.tds_types import NVarCharType
+from . import lcid
+import pytds.tz
+from .tds import (
+    SimpleLoadBalancer,
+    _TdsSocket, tds7_get_instances,
+    _create_exception_by_message,
+    output, default
+)
+from . import tds_base
+from .tds_base import (
+    Error, LoginError, DatabaseError, ProgrammingError,
+    IntegrityError, DataError, InternalError,
+    InterfaceError, TimeoutError, OperationalError,
+    NotSupportedError, Warning, ClosedConnectionError,
+    Column,
+    PreLoginEnc)
+
+from .tds_types import (
+    TableValuedParam, Binary
+)
+
+from .tds_base import (
+    ROWID, DECIMAL, STRING, BINARY, NUMBER, DATETIME, INTEGER, REAL, XML
+)
+
+from . import tls
+
+__author__ = 'Mikhail Denisenko <denisenkom@gmail.com>'
+__version__ = '1.9.0'
+
+
 def _ver_to_int(ver):
     maj, minor, rev = ver.split('.')
     return (int(maj) << 24) + (int(minor) << 16) + (int(rev) << 8)
 
-intversion = _ver_to_int(__version__)
 
-logger = logging.getLogger(__name__)
+intversion = _ver_to_int(__version__)
 
 #: Compliant with DB SIG 2.0
 apilevel = '2.0'
@@ -102,10 +117,13 @@ def recordtype_row_strategy(column_names):
     Column names that are not valid Python identifiers will be replaced
     with col<number>_
     """
-    import recordtype  # optional dependency
+    try:
+        from namedlist import namedlist as recordtype  # optional dependency
+    except ImportError:
+        from recordtype import recordtype  # optional dependency
     # replace empty column names with placeholders
     column_names = [name if is_valid_identifier(name) else 'col%s_' % idx for idx, name in enumerate(column_names)]
-    recordtype_row_class = recordtype.recordtype('Row', column_names)
+    recordtype_row_class = recordtype('Row', column_names)
 
     # custom extension class that supports indexing
     class Row(recordtype_row_class):
@@ -129,6 +147,12 @@ class Connection(object):
     def __init__(self):
         self._closed = False
         self._conn = None
+        self._isolation_level = 0
+        self._autocommit = True
+        self._row_strategy = tuple_row_strategy
+        self._login = None
+        self._use_tz = None
+        self._tzinfo_factory = None
 
     @property
     def as_dict(self):
@@ -266,10 +290,15 @@ class Connection(object):
                 except Exception as e:
                     last_error = LoginError("Cannot connect to server '{0}': {1}".format(host, e), e)
                 else:
+                    sock.setsockopt(socket.SOL_TCP, socket.TCP_NODELAY, 1)
+
+                    # default keep alive should be 30 seconds according to spec:
+                    # https://msdn.microsoft.com/en-us/library/dd341108.aspx
+                    sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 30)
+
+                    sock.settimeout(retry_time)
+                    conn = _TdsSocket(self._use_tz)
                     try:
-                        sock.setsockopt(socket.SOL_TCP, socket.TCP_NODELAY, 1)
-                        sock.settimeout(retry_time)
-                        conn = _TdsSocket(self._use_tz)
                         conn.login(login, sock, self._tzinfo_factory)
                         if conn.mars_enabled:
                             cursor = _MarsCursor(
@@ -359,8 +388,8 @@ class Connection(object):
                     return _MarsCursor(self,
                                        self._conn.create_session(self._tzinfo_factory),
                                        self._tzinfo_factory)
-                except socket.error as e:
-                    if e.errno != errno.ECONNRESET:
+                except (socket.error, OSError) as e:
+                    if e.errno not in (errno.EPIPE, errno.ECONNRESET):
                         raise
                 except ClosedConnectionError:
                     pass
@@ -421,7 +450,7 @@ class Connection(object):
             if session.in_cancel:
                 session.process_cancel()
 
-            if session.state == TDS_PENDING:
+            if session.state == tds_base.TDS_PENDING:
                 raise InterfaceError('Results are still pending on connection')
             self._active_cursor = cursor
 
@@ -574,10 +603,10 @@ class Cursor(six.Iterator):
         self._ensure_transaction()
         operation = six.text_type(operation)
         if params:
+            named_params = {}
             if isinstance(params, (list, tuple)):
                 names = []
                 pid = 1
-                named_params = {}
                 for val in params:
                     if val is None:
                         names.append('NULL')
@@ -593,7 +622,6 @@ class Cursor(six.Iterator):
             elif isinstance(params, dict):
                 # prepend names with @
                 rename = {}
-                named_params = {}
                 for name, value in params.items():
                     if value is None:
                         rename[name] = 'NULL'
@@ -608,7 +636,7 @@ class Cursor(six.Iterator):
                     u'{0} {1}'.format(p.column_name, p.type.get_declaration())
                     for p in named_params)
                 self._exec_with_retry(lambda: self._session.submit_rpc(
-                    SP_EXECUTESQL,
+                    tds_base.SP_EXECUTESQL,
                     [self._session.make_param('', operation), self._session.make_param('', param_definition)] + named_params,
                     0))
             else:
@@ -716,7 +744,6 @@ class Cursor(six.Iterator):
     def messages(self):
         """ Messages generated by server, see http://legacy.python.org/dev/peps/pep-0249/#cursor-messages
         """
-        #warnings.warn('DB-API extension cursor.messages used')
         if self._session:
             result = []
             for msg in self._session.messages:
@@ -755,7 +782,7 @@ class Cursor(six.Iterator):
             size = self.arraysize
 
         rows = []
-        for i in xrange(size):
+        for _ in xrange(size):
             row = self.fetchone()
             if not row:
                 break
@@ -773,13 +800,15 @@ class Cursor(six.Iterator):
             raise StopIteration
         return row
 
-    def setinputsizes(self, sizes=None):
+    @staticmethod
+    def setinputsizes(sizes=None):
         """
         This method does nothing, as permitted by DB-API specification.
         """
         pass
 
-    def setoutputsize(self, size=None, column=0):
+    @staticmethod
+    def setoutputsize(size=None, column=0):
         """
         This method does nothing, as permitted by DB-API specification.
         """
@@ -787,7 +816,8 @@ class Cursor(six.Iterator):
 
     def copy_to(self, file, table_or_view, sep='\t', columns=None,
                 check_constraints=False, fire_triggers=False, keep_nulls=False,
-                kb_per_batch=None, rows_per_batch=None, order=None, tablock=False):
+                kb_per_batch=None, rows_per_batch=None, order=None, tablock=False,
+                schema=None, null_string=None):
         """ *Experimental*. Efficiently load data to database from file using ``BULK INSERT`` operation
 
         :param file: Source file-like object, should be in csv format
@@ -819,14 +849,29 @@ class Cursor(six.Iterator):
           Can be used to optimize performance, see MSSQL server documentation for details
         :type order: list
         :keyword tablock: Enable or disable table lock for the duration of bulk load
+        :keyword schema: Name of schema for table or view, if not specified default schema will be used
+        :keyword null_string: String that should be interpreted as a NULL when reading the CSV file.
         """
         conn = self._conn()
         import csv
         reader = csv.reader(file, delimiter=sep)
-        if not columns:
-            self.execute('select top 1 * from [{}] where 1<>1'.format(table_or_view))
-            columns = [col[0] for col in self.description]
-        metadata = [Column(name=col, type=conn._conn.NVarChar(4000), flags=Column.fNullable) for col in columns]
+
+        if null_string is not None:
+            def _convert_null_strings(csv_reader):
+                for row in csv_reader:
+                    yield [r if r != null_string else None for r in row]
+
+            reader = _convert_null_strings(reader)
+
+        obj_name = tds_base.tds_quote_id(table_or_view)
+        if schema:
+            obj_name = '{0}.{1}'.format(tds_base.tds_quote_id(schema), obj_name)
+        if columns:
+            metadata = [Column(name=name, type=NVarCharType(size=4000), flags=Column.fNullable) for name in columns]
+        else:
+            self.execute('select top 1 * from {} where 1<>1'.format(obj_name))
+            metadata = [Column(name=col[0], type=NVarCharType(size=4000), flags=Column.fNullable if col[6] else 0)
+                        for col in self.description]
         col_defs = ','.join('{0} {1}'.format(col.column_name, col.type.get_declaration())
                             for col in metadata)
         with_opts = []
@@ -847,7 +892,7 @@ class Cursor(six.Iterator):
         with_part = ''
         if with_opts:
             with_part = 'WITH ({0})'.format(','.join(with_opts))
-        operation = 'INSERT BULK [{0}]({1}) {2}'.format(table_or_view, col_defs, with_part)
+        operation = 'INSERT BULK {0}({1}) {2}'.format(obj_name, col_defs, with_part)
         self.execute(operation)
         self._session.submit_bulk(metadata, reader)
         self._session.process_simple_request()
@@ -988,11 +1033,13 @@ def _parse_connection_string(connstr):
 
 def connect(dsn=None, database=None, user=None, password=None, timeout=None,
             login_timeout=15, as_dict=None,
-            appname=None, port=None, tds_version=TDS74,
-            encryption_level=TDS_ENCRYPTION_OFF, autocommit=False,
+            appname=None, port=None, tds_version=tds_base.TDS74,
+            autocommit=False,
             blocksize=4096, use_mars=False, auth=None, readonly=False,
             load_balancer=None, use_tz=None, bytes_to_unicode=True,
-            row_strategy=None, failover_partner=None, server=None):
+            row_strategy=None, failover_partner=None, server=None,
+            cafile=None, validate_host=True, enc_login_only=False,
+            ):
     """
     Opens connection to the database
 
@@ -1018,7 +1065,6 @@ def connect(dsn=None, database=None, user=None, password=None, timeout=None,
     :type port: int
     :keyword tds_version: Maximum TDS version to use, should only be used for testing
     :type tds_version: int
-    :keyword encryption_level: Encryption level to use, not supported
     :keyword autocommit: Enable or disable database level autocommit
     :type autocommit: bool
     :keyword blocksize: Size of block for the TDS protocol, usually should not be used
@@ -1039,12 +1085,20 @@ def connect(dsn=None, database=None, user=None, password=None, timeout=None,
       :func:`tuple_row_strategy`, :func:`list_row_strategy`, :func:`dict_row_strategy`,
       :func:`namedtuple_row_strategy`, :func:`recordtype_row_strategy`
     :type row_strategy: function of list of column names returning row factory
+    :keyword cafile: Name of the file containing trusted CAs in PEM format, if provided will enable TLS
+    :type cafile: str
+    :keyword validate_host: Host name validation during TLS connection is enabled by default, if you disable it you
+      will be vulnerable to MitM type of attack.
+    :type validate_host: bool
+    :keyword enc_login_only: Allows you to scope TLS encryption only to an authentication portion.  This means that
+      anyone who can observe traffic on your network will be able to see all your SQL requests and potentially modify
+      them.
+    :type enc_login_only: bool
     :returns: An instance of :class:`Connection`
     """
     login = _TdsLogin()
     login.client_host_name = socket.gethostname()[:128]
     login.library = "Python TDS Library"
-    login.encryption_level = encryption_level
     login.user_name = user or ''
     login.password = password or ''
     login.app_name = appname or 'pytds'
@@ -1060,17 +1114,33 @@ def connect(dsn=None, database=None, user=None, password=None, timeout=None,
     login.pid = os.getpid()
     login.change_password = ''
     login.client_id = uuid.getnode()  # client mac address
+    login.cafile = cafile
+    login.validate_host = validate_host
+    login.enc_login_only = enc_login_only
+    if cafile:
+        if not tls.OPENSSL_AVAILABLE:
+            raise ValueError("You are trying to use encryption but pyOpenSSL does not work, you probably "
+                             "need to install it first")
+        login.tls_ctx = tls.create_context(cafile)
+        if login.enc_login_only:
+            login.enc_flag = PreLoginEnc.ENCRYPT_OFF
+        else:
+            login.enc_flag = PreLoginEnc.ENCRYPT_ON
+    else:
+        login.tls_ctx = None
+        login.enc_flag = PreLoginEnc.ENCRYPT_NOT_SUP
+
     if use_tz:
         login.client_tz = use_tz
     else:
-        login.client_tz = tz.local
+        login.client_tz = pytds.tz.local
 
     # that will set:
     # ANSI_DEFAULTS to ON,
     # IMPLICIT_TRANSACTIONS to OFF,
     # TEXTSIZE to 0x7FFFFFFF (2GB) (TDS 7.2 and below), TEXTSIZE to infinite (introduced in TDS 7.3),
     # and ROWCOUNT to infinite
-    login.option_flag2 = TDS_ODBC_ON
+    login.option_flag2 = tds_base.TDS_ODBC_ON
 
     login.connect_timeout = login_timeout
     login.query_timeout = timeout
@@ -1110,10 +1180,11 @@ def connect(dsn=None, database=None, user=None, password=None, timeout=None,
     conn._autocommit = autocommit
     conn._login = login
 
-    assert row_strategy == None or as_dict == None, 'Both row_startegy and as_dict were specified, you should use either one or another'
-    if as_dict != None:
+    assert row_strategy is None or as_dict is None,\
+        'Both row_startegy and as_dict were specified, you should use either one or another'
+    if as_dict is not None:
         conn.as_dict = as_dict
-    elif row_strategy != None:
+    elif row_strategy is not None:
         conn._row_strategy = row_strategy
     else:
         conn._row_strategy = tuple_row_strategy # default row strategy
@@ -1127,15 +1198,15 @@ def connect(dsn=None, database=None, user=None, password=None, timeout=None,
 
 
 def Date(year, month, day):
-    return date(year, month, day)
+    return datetime.date(year, month, day)
 
 
 def DateFromTicks(ticks):
-    return date.fromtimestamp(ticks)
+    return datetime.date.fromtimestamp(ticks)
 
 
 def Time(hour, minute, second, microsecond=0, tzinfo=None):
-    return time(hour, minute, second, microsecond, tzinfo)
+    return datetime.time(hour, minute, second, microsecond, tzinfo)
 
 
 def TimeFromTicks(ticks):
@@ -1144,8 +1215,8 @@ def TimeFromTicks(ticks):
 
 
 def Timestamp(year, month, day, hour, minute, second, microseconds=0, tzinfo=None):
-    return datetime(year, month, day, hour, minute, second, microseconds, tzinfo)
+    return datetime.datetime(year, month, day, hour, minute, second, microseconds, tzinfo)
 
 
 def TimestampFromTicks(ticks):
-    return datetime.fromtimestamp(ticks)
+    return datetime.datetime.fromtimestamp(ticks)
